import typing as tp

from logging import getLogger

from .base import BaseClass

logger = getLogger(__name__)


class ReqRes(BaseClass):
    """
    Class for handling Robonomics reqres rpc requests
    """

    def p2p_get(
        self, address: str, message: str, result_handler: tp.Optional[tp.Callable] = None
    ) -> tp.Dict[str, tp.Union[str, int]]:
        """
        Returns for p2p rpc get response.

        :param address: Multiaddr address of the peer to connect to. For example:
            "/ip4/127.0.0.1/tcp/61240/<Peer ID of server>."
            This ID may be obtained on node/server initialization.
        :param message: Request message. ``GET`` for example.
        :param result_handler: Callback function that processes the result received from the node. This function accepts
            one argument - response.

        """

        return self._service_functions.rpc_request("p2p_get", [address, message], result_handler)

    def p2p_ping(
        self, address: str, result_handler: tp.Optional[tp.Callable] = None
    ) -> tp.Dict[str, tp.Union[str, int]]:

        """
        Returns for reqres p2p rpc ping to server response

        :param address: Multiaddr address of the peer to connect to. For example:
            "/ip4/127.0.0.1/tcp/61240/<Peer ID of server>."
            This ID may be obtained on node/server initialization.
        :param result_handler: Callback function that processes the result received from the node. This function accepts
            one argument - response.

        """

        return self._service_functions.rpc_request("p2p_ping", [address], result_handler)



class NoPrivateKeyException(Exception):
    """
    No private key was provided so unable to perform any operations requiring message signing.

    """

    pass


class DigitalTwinMapException(Exception):
    """
    No Digital Twin was created with this index or there is no such topic in Digital Twin map.

    """

    pass



import typing as tp

from logging import getLogger

from .base import BaseClass
from ..types import DatalogTyping

logger = getLogger(__name__)


class Datalog(BaseClass):
    """
    Class for datalog chainstate queries and extrinsic executions.
    """

    def get_index(self, addr: tp.Optional[str] = None, block_hash: tp.Optional[str] = None) -> tp.Dict[str, int]:
        """
        Get account datalog index dictionary.

        :param addr: ss58 type ``32`` address of an account which datalog index is to be obtained. If ``None``, tries to
            get Account datalog index if keypair was created, else raises ``NoPrivateKey``.
        :param block_hash: Retrieves data as of passed block hash.

        :return: Tuple of form {'start': <int>, 'end': <int>}

        """

        address: str = addr or self.account.get_address()

        logger.info(f"Fetching datalog index of {address}")

        return self._service_functions.chainstate_query("Datalog", "DatalogIndex", address, block_hash=block_hash)

    def get_item(
        self, addr: tp.Optional[str] = None, index: tp.Optional[int] = None, block_hash: tp.Optional[str] = None
    ) -> tp.Optional[DatalogTyping]:
        """
        Fetch datalog record of a provided account. Fetch self datalog if no address provided and interface was
        initialized with a seed.

        :param addr: ss58 type ``32`` address of an account which datalog is to be fetched. If ``None``, tries to fetch
            self datalog if keypair was created, else raises ``NoPrivateKey``.
        :param index: record index. case ``int``: fetch datalog by specified index case ``None``: fetch latest datalog.
        :param block_hash: Retrieves data as of passed block hash.

        :return: Tuple. Datalog of the account with a timestamp, ``None`` if no records.

        """

        address: str = addr or self.account.get_address()

        logger.info(
            f"Fetching {'latest datalog record' if not index else 'datalog record #' + str(index)}" f" of {address}."
        )

        if index:
            record: DatalogTyping = self._service_functions.chainstate_query(
                "Datalog", "DatalogItem", [address, index], block_hash=block_hash
            )
            return record if record[0] != 0 else None
        else:
            index_latest: int = self.get_index(address)["end"] - 1
            return (
                self._service_functions.chainstate_query(
                    "Datalog", "DatalogItem", [address, index_latest], block_hash=block_hash
                )
                if index_latest != -1
                else None
            )

    def record(self, data: str, nonce: tp.Optional[int] = None) -> str:
        """
        Write any string to datalog. It has 512 bytes length limit.

        :param data: String to be stored in datalog. It has 512 bytes length limit.
        :param nonce: Nonce of the transaction. Due to the feature of substrate-interface lib, to create an extrinsic
            with incremented nonce, pass account's current nonce. See
            https://github.com/polkascan/py-substrate-interface/blob/85a52b1c8f22e81277907f82d807210747c6c583/substrateinterface/base.py#L1535
            for example.

        :return: Hash of the datalog record transaction.

        """

        logger.info(f"Writing datalog {data}")
        return self._service_functions.extrinsic("Datalog", "record", {"record": data}, nonce)

    def erase(self, nonce: tp.Optional[int] = None) -> str:
        """
        Erase ALL datalog records of Account.

        :param nonce: Nonce of the transaction. Due to the feature of substrate-interface lib, to create an extrinsic
            with incremented nonce, pass account's current nonce. See
            https://github.com/polkascan/py-substrate-interface/blob/85a52b1c8f22e81277907f82d807210747c6c583/substrateinterface/base.py#L1535
            for example.

        :return: Hash of the datalog erase transaction.

        """

        logger.info(f"Erasing all datalogs of Account")
        return self._service_functions.extrinsic("Datalog", "erase", nonce)



from .account import Account
from .base import BaseClass
from .common_functions import CommonFunctions
from .datalog import Datalog
from .digital_twin import DigitalTwin
from .launch import Launch
from .liability import Liability
from .pubsub import PubSub
from .reqres import ReqRes
from .rws import RWS
from .service_functions import ServiceFunctions
from .subscriptions import SubEvent, Subscriber



import typing as tp

from logging import getLogger

from .base import BaseClass
from ..types import ListenersResponse

logger = getLogger(__name__)


class PubSub(BaseClass):
    """
    Class for handling Robonomics pubsub rpc requests

    WARNING: THIS MODULE IS UNDER CONSTRUCTION, USE AT YOUR OWN RISK! TO BE UPDATED SOON
    """

    def connect(
        self, address: str, result_handler: tp.Optional[tp.Callable] = None
    ) -> tp.Dict[str, tp.Union[str, bool, int]]:
        """
        Connect to peer and add it into swarm.

        :param address: Multiaddr address of the peer to connect to.
        :param result_handler: Callback function that processes the result received from the node.

        :return: Success flag in JSON message.

        """

        return self._service_functions.rpc_request("pubsub_connect", [address], result_handler)

    def listen(
        self, address: str, result_handler: tp.Optional[tp.Callable] = None
    ) -> tp.Dict[str, tp.Union[str, bool, int]]:
        """
        Listen address for incoming connections.

        :param address: Multiaddr address of the peer to connect to.
        :param result_handler: Callback function that processes the result received from the node.

        :return: Success flag in JSON message.

        """

        return self._service_functions.rpc_request("pubsub_listen", [address], result_handler)

    def get_listeners(self, result_handler: tp.Optional[tp.Callable] = None) -> ListenersResponse:
        """
        Returns a list of node addresses.

        :param result_handler: Callback function that processes the result received from the node.

        :return: List of node addresses in JSON message.

        """

        return self._service_functions.rpc_request("pubsub_listeners", None, result_handler)

    def get_peer(self, result_handler: tp.Optional[tp.Callable] = None) -> tp.Dict[str, tp.Union[str, int]]:
        """
        Returns local peer ID.

        :return: Local peer ID in JSON message.

        """

        return self._service_functions.rpc_request("pubsub_peer", None, result_handler)

    def publish(self, topic_name: str, message: str, result_handler: tp.Optional[tp.Callable] = None) -> tp.Any:
        """
        Publish message into the topic by name.

        :param topic_name: Topic name.
        :param message: Message to be published.
        :param result_handler: Callback function that processes the result received from the node.

        :return: TODO

        """

        return self._service_functions.rpc_request("pubsub_publish", [topic_name, message], result_handler)

    def subscribe(
        self, topic_name: str, result_handler: tp.Optional[tp.Callable] = None
    ) -> tp.Dict[str, tp.Union[str, int]]:
        """
        Listen address for incoming connections.

        :param topic_name: Topic name to subscribe to.
        :param result_handler: Callback function that processes the result received from the node.

        :return: Subscription ID in JSON message.

        """

        return self._service_functions.rpc_request("pubsub_subscribe", [topic_name], result_handler)

    def unsubscribe(
        self, subscription_id: str, result_handler: tp.Optional[tp.Callable] = None
    ) -> tp.Dict[str, tp.Union[str, bool, int]]:
        """
        Unsubscribe for incoming messages from topic.

        :param subscription_id: Subscription ID obtained when subscribed.
        :param result_handler: Callback function that processes the result received from the node.

        :return: Success flag in JSON message.

        """

        return self._service_functions.rpc_request("pubsub_unsubscribe", [subscription_id], result_handler)



from .constants import *
from .exceptions import *
from .types import *
from .utils import *
from .classes import (
    BaseClass,
    Account,
    CommonFunctions,
    Datalog,
    DigitalTwin,
    Launch,
    Liability,
    PubSub,
    ReqRes,
    RWS,
    ServiceFunctions,
    SubEvent,
    Subscriber,
)



import typing as tp

from logging import getLogger
from scalecodec.types import GenericCall, GenericExtrinsic
from substrateinterface import Keypair, SubstrateInterface, ExtrinsicReceipt
from substrateinterface.exceptions import ExtrinsicFailedException

from .account import Account
from ..decorators import check_socket_opened
from ..types import QueryParams, TypeRegistryTyping, RWSParamsTyping
from ..exceptions import NoPrivateKeyException

logger = getLogger(__name__)


class ServiceFunctions:
    """
    Class for custom queries, extrinsics and RPC calls to Robonomics parachain network.
    """

    def __init__(self, account: Account, wait_for_inclusion: bool = True, rws_sub_owner: tp.Optional[str] = None):
        """
        Assign Account dataclass parameters and create an empty interface attribute for a decorator.

        :param account: Account dataclass with ``seed``, ``remote_ws`` and node ``type_registry``.
        :param wait_for_inclusion: Whether wait for a transaction to included in block. You will get the hash anyway.
        :param rws_sub_owner: Subscription owner address. If passed, all extrinsics will be executed via RWS
            subscriptions.

        """
        self.remote_ws: str = account.remote_ws
        self.type_registry: TypeRegistryTyping = account.type_registry
        self.keypair: Keypair = account.keypair
        self.interface: tp.Optional[SubstrateInterface] = None
        self.wait_for_inclusion: bool = wait_for_inclusion
        self.rws_sub_owner: tp.Optional[str] = rws_sub_owner

    @check_socket_opened
    def chainstate_query(
        self,
        module: str,
        storage_function: str,
        params: QueryParams = None,
        block_hash: tp.Optional[str] = None,
        subscription_handler: tp.Optional[callable] = None,
    ) -> tp.Any:
        """
        Create custom queries to fetch data from the Chainstate. Module names and storage functions, as well as required
        parameters are available at https://parachain.robonomics.network/#/chainstate.

        :param module: Chainstate module.
        :param storage_function: Storage function.
        :param params: Query parameters. None if no parameters. Include in list, if several.
        :param block_hash: Retrieves data as of passed block hash.
        :param subscription_handler: Callback function that processes the updates of the storage query subscription.
            The workflow is the same as in substrateinterface lib. Calling method with this parameter blocks current
            thread! Example of subscription handler:
            https://github.com/polkascan/py-substrate-interface#storage-subscriptions

        :return: Output of the query in any form.

        """

        logger.info(f"Performing query {module}.{storage_function}")
        return self.interface.query(
            module,
            storage_function,
            [params] if params is not None else None,
            block_hash=block_hash,
            subscription_handler=subscription_handler,
        ).value

    @check_socket_opened
    def extrinsic(
        self,
        call_module: str,
        call_function: str,
        params: tp.Optional[tp.Dict[str, tp.Any]] = None,
        nonce: tp.Optional[int] = None,
    ) -> str:
        """
        Create an extrinsic, sign&submit it. Module names and functions, as well as required parameters are available
        at https://parachain.robonomics.network/#/extrinsics.

        :param call_module: Call module from extrinsic tab on portal.
        :param call_function: Call function from extrinsic tab on portal.
        :param params: Call parameters as a dictionary. ``None`` for no parameters.
        :param nonce: Transaction nonce, defined automatically if None. Due to the feature of substrate-interface lib,
            to create an extrinsic with incremented nonce, pass account's current nonce. See
            https://github.com/polkascan/py-substrate-interface/blob/85a52b1c8f22e81277907f82d807210747c6c583/substrateinterface/base.py#L1535
            for example.

        :return: Extrinsic hash.

        """

        if not self.keypair:
            raise NoPrivateKeyException("No seed was provided, unable to use extrinsics.")

        if not self.rws_sub_owner:
            logger.info(f"Creating a call {call_module}:{call_function}")
            call: GenericCall = self.interface.compose_call(
                call_module=call_module, call_function=call_function, call_params=params or None
            )
        else:
            logger.info(f"Creating an RWS call {call_module}:{call_function}")

            rws_params: RWSParamsTyping = {
                "subscription_id": self.rws_sub_owner,
                "call": {"call_module": call_module, "call_function": call_function, "call_args": params},
            }

            call: GenericCall = self.interface.compose_call(
                call_module="RWS", call_function="call", call_params=rws_params
            )

        logger.info("Creating extrinsic")
        extrinsic: GenericExtrinsic = self.interface.create_signed_extrinsic(
            call=call, keypair=self.keypair, nonce=nonce
        )

        logger.info("Submitting extrinsic")
        receipt: ExtrinsicReceipt = self.interface.submit_extrinsic(
            extrinsic, wait_for_inclusion=self.wait_for_inclusion
        )

        if self.wait_for_inclusion is True and not receipt.is_success:
            raise ExtrinsicFailedException()

        logger.info(
            f"Extrinsic {receipt.extrinsic_hash} for RPC {call_module}:{call_function} submitted and "
            f"included in block {receipt.block_hash}"
        )

        return str(receipt.extrinsic_hash)

    @check_socket_opened
    def rpc_request(
        self, method: str, params: tp.Optional[tp.List[str]], result_handler: tp.Optional[tp.Callable]
    ) -> tp.Dict[str, tp.Any]:
        """
        Method that handles the actual RPC request to the Substrate node. The other implemented functions eventually
        use this method to perform the request.

        :param method: Method of the ``JSONRPC`` request.
        :param params: A list containing the parameters of the ``JSONRPC`` request.
        :param result_handler: Callback function that processes the result received from the node.

        :return: Result of the request.

        """

        return self.interface.rpc_request(method, params, result_handler)

    @check_socket_opened
    def subscribe_block_headers(self, callback: callable) -> dict:
        """
        Get chain head block headers.

        :return: Chain head block headers.

        """

        return self.interface.subscribe_block_headers(subscription_handler=callback)



import typing as tp

from logging import getLogger
from scalecodec.base import ScaleBytes

from .base import BaseClass
from ..exceptions import NoPrivateKeyException
from ..types import LiabilityTyping, ReportTyping
from ..utils import ipfs_qm_hash_to_32_bytes, str_to_scalebytes

logger = getLogger(__name__)


class Liability(BaseClass):
    """
    Class for interacting with Robonomics Liability. Create and finalize ones, get information.
    """

    def get_agreement(self, index: int, block_hash: tp.Optional[str] = None) -> tp.Optional[LiabilityTyping]:
        """
        Fetch information about existing liabilities.

        :param index: Liability item index.
        :param block_hash: Retrieves data as of passed block hash.

        :return: Liability information: ``technics``, ``economics``, ``promisee``, ``promisor``, ``signatures``.
            ``None`` if no such liability.

        """
        logger.info(f"Fetching information about liability with index {index}")

        return self._service_functions.chainstate_query("Liability", "AgreementOf", index, block_hash=block_hash)

    def get_latest_index(self, block_hash: tp.Optional[str] = None) -> tp.Optional[int]:
        """
        Fetch total number of liabilities in chain (method returns the latest liability index +1).

        :param block_hash: Retrieves data as of passed block hash.

        :return: Total number of liabilities in chain (method returns the latest liability index +1). None if no
            liabilities.

        """

        logger.info("Fetching total number of liabilities in chain.")

        return self._service_functions.chainstate_query("Liability", "LatestIndex", block_hash=block_hash)

    def get_report(self, index: int, block_hash: tp.Optional[str] = None) -> tp.Optional[ReportTyping]:
        """
        Fetch information about existing liability reports.

        :param index: Reported liability item index.
        :param block_hash: block_hash: Retrieves data as of passed block hash.

        :return: Liability report information: ``index``, ``promisor``, ``report``, ``signature``. ``None`` if no such
            liability report.

        """

        logger.info(f"Fetching information about reported liability with index {index}")

        return self._service_functions.chainstate_query("Liability", "ReportOf", index, block_hash=block_hash)

    def create(
        self,
        technics_hash: str,
        economics: int,
        promisee: str,
        promisor: str,
        promisee_params_signature: str,
        promisor_params_signature: str,
        nonce: tp.Optional[int] = None,
    ) -> tp.Tuple[int, str]:
        """
        Create a liability to ensure economical relationships between robots! This is a contract to be assigned to a
        ``promisor`` by ``promisee``. As soon as the job is done and reported, the ``promisor`` gets his reward.
        This extrinsic may be submitted by another address, but there should be ``promisee`` and ``promisor``
        signatures.

        :param technics_hash: Details of the liability, where the ``promisee`` order is described.
            Accepts any 32-bytes data or a base58 (``Qm...``) IPFS hash.
        :param economics: ``Promisor`` reward in Weiners.
        :param promisee: ``Promisee`` (customer) ss58 address
        :param promisor: ``Promisor`` (worker) ss58 address
        :param promisee_params_signature: An agreement proof. This is a private key signed message containing
            ``technics`` and ``economics``. Both sides need to do this. Signed by ``promisee``.
        :param promisor_params_signature: An agreement proof. This is a private key signed message containing
            ``technics`` and ``economics``. Both sides need to do this. Signed by ``promisor``.
        :param nonce: Account nonce. Due to the feature of substrate-interface lib, to create an extrinsic with
            incremented nonce, pass account's current nonce. See
            https://github.com/polkascan/py-substrate-interface/blob/85a52b1c8f22e81277907f82d807210747c6c583/substrateinterface/base.py#L1535
            for example.

        :return: New liability index and hash of the liability creation transaction.

        """

        logger.info(
            f"Creating new liability with promisee {promisee}, promisor {promisor}, technics {technics_hash} and"
            f"economics {economics}."
        )

        if technics_hash.startswith("Qm"):
            technics_hash = ipfs_qm_hash_to_32_bytes(technics_hash)

        liability_creation_transaction_hash: str = self._service_functions.extrinsic(
            "Liability",
            "create",
            {
                "agreement": {
                    "technics": {"hash": technics_hash},
                    "economics": {"price": economics},
                    "promisee": promisee,
                    "promisor": promisor,
                    "promisee_signature": {"Sr25519": promisee_params_signature},
                    "promisor_signature": {"Sr25519": promisor_params_signature},
                }
            },
            nonce=nonce,
        )

        liability_total: int = self.get_latest_index()
        index: int = liability_total - 1
        for liabilities in reversed(range(liability_total)):
            if self.get_agreement(liabilities)["promisee_signature"]["Sr25519"] == promisee_params_signature:
                index = liabilities
                break

        return index, liability_creation_transaction_hash

    def sign_liability(self, technics_hash: str, economics: int) -> str:
        """
        Sign liability params approve message with a private key. This function is meant to sign ``technics`` and
        ``economics``details message to state the agreement of ``promisee`` and ``promisor``. Both sides need to do this.

        :param technics_hash: Details of the liability, where the ``promisee`` order is described.
            Accepts any 32-bytes data or a base58 (``Qm...``) IPFS hash.
        :param economics: ``Promisor`` reward in Weiners.

        :return: Signed message 64-byte hash in sting form.

        """

        if not self.account.keypair:
            raise NoPrivateKeyException("No private key, unable to sign a liability")

        if technics_hash.startswith("Qm"):
            technics_hash = ipfs_qm_hash_to_32_bytes(technics_hash)

        logger.info(f"Signing proof with technics {technics_hash} and economics {economics}.")

        signed_data: ScaleBytes = str_to_scalebytes(technics_hash, "H256") + str_to_scalebytes(
            economics, "Compact<Balance>"
        )

        return f"0x{self.account.keypair.sign(signed_data).hex()}"

    def finalize(
        self,
        index: int,
        report_hash: str,
        promisor: tp.Optional[str] = None,
        promisor_finalize_signature: tp.Optional[str] = None,
        nonce: tp.Optional[int] = None,
    ) -> str:
        """
        Report on a completed job to receive a deserved award. This may be done by another address, but there should be
        a liability ``promisor`` signature.

        :param index: Liability item index.
        :param report_hash: IPFS hash of a report data (videos, text, etc.). Accepts any 32-bytes data or a base58
            (``Qm...``) IPFS hash.
        :param promisor: ``Promisor`` (worker) ss58 address. If not passed, replaced with transaction author address.
        :param promisor_finalize_signature: 'Job done' proof. A message containing liability index and report data
            signed by ``promisor``. If not passed, this message is signed by a transaction author which should be a
            ``promisor`` so.
        :param nonce: Account nonce. Due to the feature of substrate-interface lib, to create an extrinsic with
            incremented nonce, pass account's current nonce. See
            https://github.com/polkascan/py-substrate-interface/blob/85a52b1c8f22e81277907f82d807210747c6c583/substrateinterface/base.py#L1535
            for example.

        :return: Liability finalization transaction hash

        """

        logger.info(f"Finalizing liability {index} by promisor {promisor or self.account.get_address()}.")

        if report_hash.startswith("Qm"):
            report_hash = ipfs_qm_hash_to_32_bytes(report_hash)

        return self._service_functions.extrinsic(
            "Liability",
            "finalize",
            {
                "report": {
                    "index": index,
                    "sender": promisor or self.account.get_address(),
                    "payload": {"hash": report_hash},
                    "signature": {"Sr25519": promisor_finalize_signature or self.sign_report(index, report_hash)},
                }
            },
            nonce=nonce,
        )

    def sign_report(self, index: int, report_hash: str) -> str:
        """
        Sing liability finalization parameters proof message with a private key. This is meant to state that the job is
        done by ``promisor``.

        :param index: Liability item index.
        :param report_hash: IPFS hash of a report data (videos, text, etc.). Accepts any 32-bytes data or a base58
            (``Qm...``) IPFS hash.

        :return: Signed message 64-byte hash in sting form.

        """

        if not self.account.keypair:
            raise NoPrivateKeyException("No private key, unable to sign a report")

        if report_hash.startswith("Qm"):
            report_hash = ipfs_qm_hash_to_32_bytes(report_hash)

        logger.info(f"Signing report for liability {index} with report_hash {report_hash}.")

        signed_data: ScaleBytes = str_to_scalebytes(index, "U32") + str_to_scalebytes(report_hash, "H256")

        return f"0x{self.account.keypair.sign(signed_data).hex()}"



import typing as tp

from logging import getLogger

from .base import BaseClass
from ..exceptions import DigitalTwinMapException
from ..types import DigitalTwinTyping
from ..utils import dt_encode_topic

logger = getLogger(__name__)


class DigitalTwin(BaseClass):
    """
    Class for interacting with `Digital Twins <https://wiki.robonomics.network/docs/en/digital-twins/>`_..
    """

    def get_info(self, dt_id: int, block_hash: tp.Optional[str] = None) -> tp.Optional[DigitalTwinTyping]:
        """
        Fetch information about existing digital twin.

        :param dt_id: Digital Twin object ID.
        :param block_hash: Retrieves data as of passed block hash.

        :return: List of DigitalTwin associated mapping. ``None`` if no Digital Twin with such id.

        """
        logger.info(f"Fetching info about Digital Twin with ID {dt_id}")

        return self._service_functions.chainstate_query("DigitalTwin", "DigitalTwin", dt_id, block_hash=block_hash)

    def get_owner(self, dt_id: int, block_hash: tp.Optional[str] = None) -> tp.Optional[str]:
        """
        Fetch existing Digital Twin owner address.

        :param dt_id: Digital Twin object ID.
        :param block_hash: Retrieves data as of passed block hash.

        :return: Owner address. ``None`` if no Digital Twin with such id.

        """

        logger.info(f"Fetching owner of Digital Twin with ID {dt_id}")

        return self._service_functions.chainstate_query("DigitalTwin", "Owner", dt_id, block_hash=block_hash)

    def get_total(self, block_hash: tp.Optional[str] = None) -> tp.Optional[int]:
        """
        Fetch total number of Digital Twins.

        :param block_hash: Retrieves data as of passed block hash.

        :return: Total number of Digital Twins. ``None`` if no Digital Twins.

        """
        logger.info("Fetching Total number of Digital Twins")

        return self._service_functions.chainstate_query("DigitalTwin", "Total", block_hash=block_hash)

    def get_source(self, dt_id: int, topic: str, block_hash: tp.Optional[str] = None) -> str:
        """
        Find a source for a passed Digital Twin topic.

        :param dt_id: Digital Twin id.
        :param topic: Searched topic. Normal string.
        :param block_hash: Retrieves data as of passed block hash.

        :return: If found, topic source ss58 address.

        """

        dt_map: tp.Optional[DigitalTwinTyping] = self.get_info(dt_id, block_hash=block_hash)
        if not dt_map:
            raise DigitalTwinMapException("No Digital Twin was created or Digital Twin map is empty.")
        topic_hashed: str = dt_encode_topic(topic)
        for source in dt_map:
            if source[0] == topic_hashed:
                return source[1]
        raise DigitalTwinMapException(f"No topic {topic} was found in Digital Twin with id {dt_id}")

    def create(self, nonce: tp.Optional[int] = None) -> tp.Tuple[int, str]:
        """
        Create a new digital twin.

        :param nonce: Account nonce. Due to the feature of substrate-interface lib, to create an extrinsic with
            incremented nonce, pass account's current nonce. See
            https://github.com/polkascan/py-substrate-interface/blob/85a52b1c8f22e81277907f82d807210747c6c583/substrateinterface/base.py#L1535
            for example.

        :return: Tuple of newly created Digital Twin ID and hash of the creation transaction.

        """

        tr_hash: str = self._service_functions.extrinsic("DigitalTwin", "create", nonce=nonce)
        dt_total: int = self.get_total()
        dt_id: int = dt_total
        for ids in reversed(range(dt_total)):
            if self.get_owner(ids) == self.account.get_address():
                dt_id: int = ids
                break

        return dt_id, tr_hash

    def set_source(self, dt_id: int, topic: str, source: str, nonce: tp.Optional[int] = None) -> tp.Tuple[str, str]:
        """
        Set DT topics and their sources. Since ``topic_name`` is byte encoded and then sha256-hashed, it's considered as
        good practice saving the map of digital twin in human-readable format in the very first DT topic. Still there is
        a ``get_source`` function which transforms given string to the format as saved in the chain for comparing.

        :param dt_id: Digital Twin ID, which should have been created by account, calling this function.
        :param topic: Topic to add. Any string you want. It will be sha256 hashed and stored in blockchain.
        :param source: Source address in ss58 format.
        :param nonce: Account nonce. Due to the feature of substrate-interface lib, to create an extrinsic with
            incremented nonce, pass account's current nonce. See
            https://github.com/polkascan/py-substrate-interface/blob/85a52b1c8f22e81277907f82d807210747c6c583/substrateinterface/base.py#L1535
            for example.

        :return: Tuple of hashed topic and transaction hash.

        """

        topic_hashed = dt_encode_topic(topic)
        return (
            topic_hashed,
            self._service_functions.extrinsic(
                "DigitalTwin", "set_source", {"id": dt_id, "topic": topic_hashed, "source": source}, nonce=nonce
            ),
        )



import typing as tp

from logging import getLogger

from .account import Account
from .service_functions import ServiceFunctions

logger = getLogger(__name__)


class BaseClass:
    """
    Base class for different modules to initialize `service_functions` instance for further work.
    """

    def __init__(self, account: Account, wait_for_inclusion: bool = True, rws_sub_owner: tp.Optional[str] = None):
        """
        Assign Account dataclass parameters and create an empty interface attribute for a decorator.

        :param account: Account dataclass with seed, websocket address and node type_registry.
        :param wait_for_inclusion: Whether wait for a transaction to included in block. You will get the hash anyway.
        :param rws_sub_owner: Subscription owner address. If passed, all extrinsics will be executed via RWS
            subscriptions.

        """
        self.account: Account = account
        self._service_functions: ServiceFunctions = ServiceFunctions(
            account, wait_for_inclusion=wait_for_inclusion, rws_sub_owner=rws_sub_owner
        )



REMOTE_WS = "wss://kusama.rpc.robonomics.network"
TYPE_REGISTRY = {
    "types": {
        "Record": "Vec<u8>",
        "<T as frame_system::Config>::AccountId": "AccountId",
        "RingBufferItem": {"type": "struct", "type_mapping": [["timestamp", "Compact<u64>"], ["payload", "Vec<u8>"]]},
        "RingBufferIndex": {"type": "struct", "type_mapping": [["start", "Compact<u64>"], ["end", "Compact<u64>"]]},
    }
}



import typing as tp
from functools import partial

from enum import Enum
from logging import getLogger
from websocket import WebSocketConnectionClosedException

from .account import Account
from .service_functions import ServiceFunctions

logger = getLogger(__name__)


class SubEvent(Enum):
    NewRecord = "NewRecord"
    NewLaunch = "NewLaunch"
    Transfer = "Transfer"
    TopicChanged = "TopicChanged"
    NewDevices = "NewDevices"


class Subscriber:
    """
    Class intended for use in cases when needed to subscribe on chainstate updates/events. **Blocks current thread**!
    """

    def __init__(
        self,
        account: Account,
        subscribed_event: SubEvent,
        subscription_handler: callable,
        pass_event_id: bool = False,
        addr: tp.Optional[tp.Union[tp.List[str], str]] = None,
    ) -> None:
        """
        Initiates an instance for further use and starts a subscription for a selected action.

        :param account: Account dataclass with ``seed``, ``remote_ws`` address and node ``type_registry``.
        :param subscribed_event: Event in substrate chain to be awaited. Choose from [``NewRecord``, ``NewLaunch``,
            ``Transfer``, ``TopicChanged``, ``NewDevices``]. This parameter should be a ``SubEvent`` class attribute.
            This also requires importing this class.
        :param subscription_handler: Callback function that processes the updates of the storage.
            THIS FUNCTION IS MEANT TO ACCEPT ONLY ONE ARGUMENT (THE NEW EVENT DESCRIPTION TUPLE) by default.
            It will receive event ID as a second parameter if ``pass_event_id`` is True.
        :param pass_event_id: The ``subscription_handler`` will receive event ID as a second parameter
            if ``pass_event_id`` is True. Format is ``{block_number}-{event_idx}``.
        :param addr: ss58 type 32 address(-es) of an account(-s) which is(are) meant to be event target. If ``None``,
            will subscribe to all such events never-mind target address(-es).

        """

        self._custom_functions: ServiceFunctions = ServiceFunctions(account)
        self._event: SubEvent = subscribed_event
        self._callback: callable = subscription_handler
        self._target_address: tp.Optional[tp.Union[tp.List[str], str]] = addr
        self._pass_event_id: bool = pass_event_id

        self._subscribe_event()

    def _subscribe_event(self) -> None:
        """
        Subscribe to events targeted to a certain account (``launch``, ``transfer``). Call ``subscription_handler``
        when updated.

        """

        logger.info(f"Subscribing to event {self._event.value} for target addresses {self._target_address}")
        try:
            self._custom_functions.subscribe_block_headers(self._event_callback)
        except WebSocketConnectionClosedException:
            self._subscribe_event()

    def _event_callback(self, index_obj: tp.Any, update_nr: int, subscription_id: int) -> None:
        """
        Function, processing updates in event list storage. On update filters events to a desired account
        and passes the event description to the user-provided ``callback`` method.

        :param index_obj: Updated event list.
        :param update_nr: Update counter. Increments every new update added. Starts with ``0``.
        :param subscription_id: Subscription ID.

        """

        if update_nr == 0:
            return None

        chain_events: list = self._custom_functions.chainstate_query("System", "Events")
        for events in chain_events:
            if events["event_id"] in self._event.value:
                should_callback: bool = (self._target_address is None) or (
                    events["event"]["attributes"][
                        0
                        if self._event
                        in [
                            SubEvent.NewRecord,
                            SubEvent.TopicChanged,
                            SubEvent.NewDevices,
                        ]
                        else 1
                    ]
                    in self._target_address
                )
                if not should_callback:
                    continue
                callback = partial(self._callback, events["event"]["attributes"])
                if self._pass_event_id:
                    block_num: int = index_obj["header"]["number"]
                    event_id: str = "{}-{}".format(block_num, events["extrinsic_idx"])
                    callback = partial(callback, event_id)
                callback()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['robonomicsinterface', 'robonomicsinterface.classes']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.4,<9.0.0', 'substrate-interface>=1.1.7']

entry_points = \
{'console_scripts': ['robonomics_interface = '
                     'robonomicsinterface.robonomics_interface_io:cli']}

setup_kwargs = {
    'name': 'robonomics-interface',
    'version': '1.0.5',
    'description': 'Robonomics wrapper over https://github.com/polkascan/py-substrate-interface created to facilitate programming with Robonomics',
    'long_description': '# robonomics-interface\nThis is a simple wrapper over https://github.com/polkascan/py-substrate-interface used to facilitate writing code for applications using Robonomics.\n\nRobonomics project: https://robonomics.network/\n\nRobonomics parachain dapp: https://parachain.robonomics.network/\n\nDocumentation: https://multi-agent-io.github.io/robonomics-interface/\n_______\n# Installation \n```bash\npip3 install robonomics-interface\n```\n# Contributing\n\nWhen contributing to this repository, please first discuss the change you wish to make via issue,\nemail, or any other method with the owners of this repository before making a change. \n\n## Pull Request Process\n\n1. Install [poetry](https://python-poetry.org/docs/) \n2. Git clone the repository\n3. Install requirements with\n```bash\npoetry install\n```\nInstalling `substrate_interface` may require [Rust](https://www.rust-lang.org/tools/install) and \n[Rustup nightly](https://rust-lang.github.io/rustup/concepts/channels.html).\n\n4. Add functions/edit code/fix issues.\n5. Make a PR.\n6. ...\n7. Profit!\n\n\n## Some important rules\n- If needed, install dependencies with\n```bash\npoetry add <lib>\n```\n- Use `ReStructuredText` docstrings.\n- Respect typing annotation.\n- Add documentation. Please take in consideration that if a new class was created, add it to `docs/source/modules.rst`.\nOther functionality is better to be described in `docs/source/usage.rst`\n- Black it:\n```bash\nblack -l 120 <modified_file>\n```\n- Check how the docs look via `make html` from the `docs` folder and checking the `docs/build/html/index.html` page.\n- Do not bump version.\n- One may test the code by\n```bash\n# in project root\npoetry build\npip3 uninstall robonomcis_interface -y  #if was installed previously\npip3 install pip3 install dist/robonomics_interface-<version>-py3-none-any.whl \npython3 <testing_script>\n```',
    'author': 'Pavel Tarasov',
    'author_email': 'p040399@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Multi-Agent-io/robonomics-interface',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import typing as tp

from logging import getLogger

from .base import BaseClass
from ..utils import ipfs_qm_hash_to_32_bytes

logger = getLogger(__name__)


class Launch(BaseClass):
    """
    Class for sending launch transactions.
    """

    def launch(self, target_address: str, parameter: str, nonce: tp.Optional[int] = None) -> str:
        """
        Send Launch command to device.

        :param target_address: Device to be triggered with launch.
        :param parameter: Launch command accompanying parameter. Should be a 32 bytes data. Also, IPFS ``Qm...`` hash is
            supported. It will be transformed into a 32 bytes string without heading ``0x`` bytes.
        :param nonce: Account nonce. Due to the feature of substrate-interface lib, to create an extrinsic with
            incremented nonce, pass account's current nonce. See
            https://github.com/polkascan/py-substrate-interface/blob/85a52b1c8f22e81277907f82d807210747c6c583/substrateinterface/base.py#L1535
            for example.

        :return: Hash of the launch transaction.

        """

        logger.info(f"Sending launch command to {target_address}")

        if parameter.startswith("Qm"):
            parameter = ipfs_qm_hash_to_32_bytes(parameter)

        return self._service_functions.extrinsic(
            "Launch", "launch", {"robot": target_address, "param": parameter}, nonce
        )
